try:
    # Need to import some symbols explicitly for command line imports when using Passagemath
    from sage.all__sagemath_symbolics import Category
except ImportError:
    pass
from .hilbert_modular_group_class import HilbertModularGroup, is_HilbertModularGroup
from .pullback import HilbertPullback
from .upper_half_plane import UpperHalfPlaneProductElement, ComplexPlaneProductElement, \
    ComplexPlaneProduct, UpperHalfPlaneProduct



r"""
Helper routine to make plots for the examples.


"""
from typing import ParamSpec
from sage.functions.other import floor, ceil
from sage.geometry.polyhedron.constructor import Polyhedron
from sage.plot.all import Graphics, point2d, parametric_plot
from sage.plot.plot3d.shapes2 import point3d
from sage.repl.rich_output import get_display_manager
from sage.repl.rich_output.output_graphics3d import (OutputSceneThreejs as threejs,
                                                     OutputSceneJmol as jmol)
from sage.rings.real_mpfr import RR
from sage.calculus.var import var
from sage.modules.free_module_element import vector
from sage.plot.plot3d.implicit_plot3d import implicit_plot3d

P = ParamSpec('P')


def plot_polygon(p: Polyhedron, basis: tuple, action: str = 'show', filename: str = '',
                 inside_polygon: bool = None,
                 norm_bound: float = 1.0,
                 curve_map: tuple = None,
                 norm_plot_factor: float = 1.0, **options: P.kwargs):
    """
    Plot a polygon in 2D or 3D.

    INPUT:

    - ``p`` -- polygon
    - ``basis`` -- basis for a lattice (usually given by an ideal)
    - ``action`` -- string (default 'show') can be 'show', 'save', 'return'
    - ``filename`` -- filename to save figure to if action = 'save'
    - `inside_bounds`` -- a list of bounds within which to restrict the lattice points
                            (to e.g. only plot inside the polygon)
    - ``norm_bound`` -- float (default: `None`) if set add a plot of xy= norm_bound
    - ``curve_map`` -- matrix - if set and norm_bound is set transform the curve xy=c with matrix
    - ``norm_plot_factor`` -- float (default: '1.0') - multiply the parameters of the xy=c plot
                                                       with this factor

    EXAMPLES::

        sage: from hilbert_modgroup.all import *
        sage: H1 = HilbertModularGroup(5)
        sage: K1 = H1.base_ring().number_field()
        sage: P1 = HilbertPullback(H1)
        sage: z = UpperHalfPlaneProductElement([0.1+I,0.2+0.5*I])
        sage: p=P1._candidate_integers_sigma(z,domain='preimage',return_polyhedron=True)
        sage: norm_args = {'norm_bound': 1.0}
        sage: plot_polygon(p,K1.fractional_ideal(1).integral_basis(),action='return',**norm_args)
        Graphics object consisting of 18 graphics primitives
        sage: x = ZZ['x'].gen()
        sage: K3.<a> = NumberField(x^3-x^2-2*x+1)
        sage: H3=HilbertModularGroup(K3)
        sage: P3=HilbertPullback(H3)
        sage: z=UpperHalfPlaneProductElement([CC(0,1.0),CC(0,1.0),CC(0,1.0)])
        sage: p=P3._candidate_integers_sigma(z, domain='polytope', return_polyhedron=True)
        sage: norm_args = {'norm_bound': P3._bound_for_sigma_norm(z), 'viewer': 'jmol',
        ....:                   'curve_map': P3.basis_matrix_ideal(), 'norm_plot_factor': 1}
        sage: G = plot_polygon(p, [1,1,1], action='return', ticks=[2,2,2], polygon=False,
        ....: inside_polygon=True, xmin=-2, xmax=2, **norm_args)
        sage: type(G)
        <class 'sage.plot.plot3d.base.Graphics3dGroup'>
    """
    xmin = options.get('xmin') or floor(min(v[0] for v in p.vertices()))
    xmax = options.get('xmax') or ceil(max(v[0] for v in p.vertices()))
    ymin = options.get('ymin') or floor(min(v[1] for v in p.vertices()))
    ymax = options.get('ymax') or ceil(max(v[1] for v in p.vertices()))
    zmin = 0
    zmax = 1
    n = len(basis)
    if n == 3:
        zmin = options.get('zmin') or floor(min(v[2] for v in p.vertices()))
        zmax = options.get('zmax') or ceil(max(v[2] for v in p.vertices()))
    elif n > 3:
        raise NotImplementedError("This function is only implemented for degrees 2 and 3.")
    plot_options = {'point': False, 'line': 'black', 'fill': (0.9, 0.9, 0.9), 'wireframe': 'black',
                    'polygon': [0.9, 0.9, 0.9]}
    show_options = {'ticks': options.get('ticks', [xmax, ymax])}
    if n > 2:
        show_options['camera_position'] = options.get('camera_position', (2.3, 2.4, 2.0))
    if 'viewer' in options:
        show_options['viewer'] = options['viewer']
    # Remove unwanted options for the polot
    options = {k: v for k, v in options.items() if k in plot_options}
    if hasattr(basis[0], 'complex_embeddings'):
        labels = [r'$\varphi_1$', r'$\varphi_2$', r'$\varphi_3$'][0: n]
    else:
        labels = ['$X_1$', '$X_2$', '$X_3$'][0: n]

    plot_options.update(options)

    G = Graphics()
    G += p.plot(**plot_options)
    if n == 2:
        G = add_pts_degree2(G, basis, xmin, xmax, ymin, ymax, inside_polygon, p)
        G.set_axes_range(xmin - 1, xmax + 1, ymin - 1, ymax + 1)
        G.axes_labels(labels)

    else:
        G = add_pts_degree3(G, basis, xmin, xmax, ymin, ymax, zmin, zmax, inside_polygon,
                            curve_map, p, norm_bound)
    # Add also the norm bounds for degree 2
    if norm_bound and n == 2:
        B = curve_map or [[1, 0], [0, 1]]
        x_min = norm_bound / ymax / norm_plot_factor
        x_max = norm_plot_factor * xmax
        x = var('x')
        for (eps1, eps2) in [(-1, -1), (-1, 1), (1, -1), (1, 1)]:
            G += parametric_plot(((B[0][0] * eps1 * x + B[0][1] * eps2 * norm_bound / x),
                                  (B[1][0] * eps1 * x + B[1][1] * eps2 * norm_bound / x)),
                                 (x, x_min, x_max), zorder=1, color='black')
    if norm_bound and n == 3:
        x = var('x')
        y = var('y')
        z = var('z')
        B = curve_map or [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        x1 = B[0][0] * x + B[0][1] * y + B[0][2] * z
        y1 = B[1][0] * x + B[1][1] * y + B[1][2] * z
        z1 = B[2][0] * x + B[2][1] * y + B[2][2] * z
        xmax = xmax*norm_plot_factor
        ymax = ymax*norm_plot_factor
        zmax = zmax*norm_plot_factor
        xmin = xmin * norm_plot_factor
        ymin = ymin * norm_plot_factor
        zmin = zmin * norm_plot_factor

        G += implicit_plot3d(abs(x1*y1*z1) == norm_bound, (x, xmin, xmax),
                             (y, ymin, ymax), (z, zmin, zmax), color='gray',
                             opacity=0.4, plot_points=100)
    return return_show(G, action, filename, **show_options)


def return_show(G: Graphics, action: str, filename: str, **show_options: P.kwargs):
    dm = get_display_manager()
    show_3d = threejs in dm.supported_output() or jmol in dm.supported_output()
    if action == 'show' and hasattr(G, 'matplotlib'):
        G.matplotlib(**show_options)
    if action == 'show' and show_3d:
        G.show(**show_options)
    elif action == 'show':
        return draw_3dplot_as_png(G, **show_options)
    elif action == 'save' and filename:
        G.save(filename, **show_options)
    elif action == 'return':
        return G
    else:
        raise ValueError(f"action: `{action}` is not a valid value.")


def draw_3dplot_as_png(G: Graphics, **show_options: P.kwargs):
    """
    Use function taken from sage_docbuild/conf.py to plot 3d plot as an image
    when no appropriate backend is available.

    EXAMPLES::

        sage: from hilbert_modgroup.all import *
        sage: x = ZZ['x'].gen()
        sage: K3.<a> = NumberField(x^3-x^2-2*x+1)
        sage: H3=HilbertModularGroup(K3)
        sage: P3=HilbertPullback(H3)
        sage: z=UpperHalfPlaneProductElement([CC(0,1.0),CC(0,1.0),CC(0,1.0)])
        sage: p=P3._candidate_integers_sigma(z,domain='polytope', return_polyhedron=True)
        sage: norm_args = {'norm_bound': P3._bound_for_sigma_norm(z),
        ....:                   'curve_map': P3.basis_matrix_ideal(), 'norm_plot_factor': 1}
        sage: G = plot_polygon(p, [1,1,1], action='return', ticks=[2,2,2], polygon=False,
        ....: inside_polygon=True, xmin=-2, xmax=2, **norm_args) # implicit doctest
        sage: type(G)
        <class 'sage.plot.plot3d.base.Graphics3dGroup'>
    """
    import matplotlib as mpl
    import matplotlib.image as mpimg
    import matplotlib.pyplot as plt
    from sage.misc.temporary_file import tmp_filename
    mpl.rcParams['image.interpolation'] = 'bilinear'
    mpl.rcParams['image.resample'] = False
    mpl.rcParams['figure.figsize'] = [8.0, 6.0]
    mpl.rcParams['figure.dpi'] = 80
    mpl.rcParams['savefig.dpi'] = 100
    fn = tmp_filename(ext=".png")
    G.save(fn, **show_options)
    img = mpimg.imread(fn)
    show_options.pop('ticks')
    plt.imshow(img)


def add_pts_degree2(G: Graphics, basis: tuple, xmin: float, xmax: float, ymin: float, ymax: float,
                    inside_polygon: bool, p: Polyhedron) -> Graphics:
    """

        Add points to the plot G for the case of degree 2.

        INPUT:
        - ``G`` -- Graphics object to add points to
        - ``basis`` -- basis for a lattice (usually given by an ideal)
        - ``xmin`` -- minimum value of the x-axis
        - ``xmax`` -- maximum value of the x-axis
        - ``ymin`` -- minimum value of the y-axis
        - ``ymax`` -- maximum value of the y-axis
        - ``inside_polygon`` -- boolean, if True, only add points inside the polygon p
        - ``p`` -- polygon

        This function iterates over all integer coordinates within the given bounds,
        calculates the corresponding point in the lattice, and adds it to the plot
        if it is within the polygon (if specified).

        EXAMPLES::

            sage: from hilbert_modgroup.all import *
            sage: H1 = HilbertModularGroup(5)
            sage: K1 = H1.base_ring().number_field()
            sage: P1 = HilbertPullback(H1)
            sage: z = UpperHalfPlaneProductElement([0.1+I,0.2+0.5*I])
            sage: p = P1._candidate_integers_sigma(z, domain='preimage', return_polyhedron=True)
            sage: norm_args = {'norm_bound': 1.0}
            sage: G = Graphics()
            sage: basis = K1.fractional_ideal(1).integral_basis()
            sage: xmin = 0
            sage: xmax = 1
            sage: ymin = 0
            sage: ymax = 1
            sage: inside_polygon = False
            sage: add_pts_degree2(G, basis, xmin, xmax, ymin, ymax, inside_polygon, p)
            Graphics object consisting of 2 graphics primitives

        """

    a, b = basis
    for i in range(floor(xmin), ceil(xmax) + 1):
        for j in range(floor(ymin), ceil(ymax) + 1):
            x = a * i + b * j
            if hasattr(x, 'complex_embeddings'):
                coordinates = [tuple(x.complex_embeddings())]
            else:
                coordinates = [(i, j)]
            if coordinates[0][0] < xmin or coordinates[0][0] > xmax:
                continue
            if coordinates[0][1] < ymin or coordinates[0][1] > ymax:
                continue

            if inside_polygon and coordinates[0] not in p:
                continue
            G += point2d(coordinates, zorder=1, color='black', size=20)
    return G


def add_pts_degree3(G, basis: tuple, xmin: float, xmax: float, ymin: float, ymax: float,
                    zmin: float, zmax: float, inside_polygon: bool = False,
                    curve_map: tuple = None, p: Polyhedron = None,
                    norm_bound: float = 1.0) -> Graphics:
    """
    Add points to the plot G for the case of degree 3

    INPUT:

    - ``basis`` - basis of the ideal
    - ``xmin``: float - minimum value of the x-axis
    - ``xmax``: float - maximum value of the x-axis
    - ``ymin``: float - minimum value of the y-axis
    - ``ymax``: float - maximum value of the y-axis
    - ``zmin``: float - minimum value of the z-axis
    - ``zmax``: float - maximum value of the z-axis
    - ``inside_polygon`` - boolean (default=False) if True, only add points inside the polygon p
    - ``curve_map`` - basis of the curve map
    - ``p`` - polygon
    - ``norm_bound`` - norm bound

    EXAMPLES:

            sage: from hilbert_modgroup.all import *
            sage: x = ZZ['x'].gen()
            sage: K3.<a> = NumberField(x^3-x^2-2*x+1)
            sage: H3=HilbertModularGroup(K3)
            sage: P3=HilbertPullback(H3)
            sage: z=UpperHalfPlaneProductElement([CC(0,1.0),CC(0,1.0),CC(0,1.0)])
            sage: p = P3._candidate_integers_sigma(z, domain='preimage', return_polyhedron=True)
            sage: norm_args = {'norm_bound': 1.0}
            sage: G = Graphics()
            sage: basis = K3.fractional_ideal(1).integral_basis()
            sage: xmin = 0
            sage: xmax = 1
            sage: ymin = 0
            sage: ymax = 1
            sage: zmin = 0
            sage: zmax = 1
            sage: inside_polygon = False
            sage: curve_map = None
            sage: G = add_pts_degree3(G, basis, xmin, xmax, ymin, ymax, zmin, zmax)
            sage: type(G)
            <class 'sage.plot.plot3d.base.Graphics3dGroup'>
    """
    a, b, c = basis
    coordinate_as_tuple = not hasattr(a, 'complex_embeddings')

    points_3d = [tuple((a * i + b * j + c * k).complex_embeddings())
                 if not coordinate_as_tuple else (i, j, k)
                 for i in range(floor(xmin), ceil(xmax) + 1)
                 for j in range(floor(ymin), ceil(ymax) + 1)
                 for k in range(floor(zmin), ceil(zmax) + 1)
                 ]
    for coordinates in points_3d:
        if coordinates[0] < xmin or coordinates[0] > xmax:
            continue
        if coordinates[1] < ymin or coordinates[1] > ymax:
            continue
        if coordinates[2] < zmin or coordinates[2] > zmax:
            continue

        if inside_polygon and coordinates not in p:
            continue
        if not curve_map:
            norm = coordinates[0] * coordinates[1] * coordinates[2]
        else:
            B = curve_map
            emb = B * vector(RR, coordinates)
            norm = emb[0] * emb[1] * emb[2]
        if norm_bound and abs(norm) > norm_bound + 0.01:
            G += point3d([coordinates], zorder=1, color='gray', size=30, opacity=0.75)
        else:
            G += point3d([coordinates], zorder=3, color='black', size=30)
    return G



#cython: profile=True
"""
Cython versions of Pullback algorithms.

Note: These are the algorithms that needs optimizing to make it all faster.
"""
from sage.functions.other import ceil
from sage.functions.other import floor
from sage.categories.sets_cat import cartesian_product
from sage.arith.misc import gcd
from sage.structure.sage_object cimport SageObject
from hilbert_modgroup.upper_half_plane cimport UpperHalfPlaneProductElement__class
from sage.rings.number_field.number_field_element cimport NumberFieldElement

cpdef integral_coordinates_in_box(bounds):
    """
    Find all integral points inside a box in R^n.

    INPUT:
    - `bounds` -- a list of upper and lower bounds defining a parallelepiped of positive volume.

    EXAMPLES::
    
        sage: from hilbert_modgroup.pullback_cython import integral_coordinates_in_box
        sage: integral_coordinates_in_box([(-1,1),(-1,1)])
        [(-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 0), (0, 1), (1, -1), (1, 0), (1, 1)]        
        sage: integral_coordinates_in_box([(-1,1),(-1.5,1.5)])
        [(-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 0), (0, 1), (1, -1), (1, 0), (1, 1)]        
        sage: integral_coordinates_in_box([(-1,1),(-1,1),(-0.5,0.5)])
        [(-1, -1, 0),
         (-1, 0, 0),
         (-1, 1, 0),
         (0, -1, 0),
         (0, 0, 0),
         (0, 1, 0),
         (1, -1, 0),
         (1, 0, 0),
         (1, 1, 0)]
        sage: integral_coordinates_in_box([(-1,1),(-1,1),(-0.5,0.5)])
        [(-1, -1, 0),
         (-1, 0, 0),
         (-1, 1, 0),
         (0, -1, 0),
         (0, 0, 0),
         (0, 1, 0),
         (1, -1, 0),
         (1, 0, 0),
         (1, 1, 0)]
    """
    if not isinstance(bounds, list):
        raise ValueError
    n = len(bounds)
    coordinates = []
    for i in range(n):
        lower = ceil(bounds[i][0])
        upper = floor(bounds[i][1])
        if lower > upper:
            raise ValueError("Bounds must give interval of positive length.")
        coordinates.append(range(lower, upper + 1))
    # result = [tuple(x) for x in cartesian_product(coordinates)]
    # return result
    return list(cartesian_product(coordinates))

cpdef lattice_elements_in_box(lattice_basis, lattice_bounds, coordinate_bounds, norm_bound=None):
    """
    Return all coordinates of elements in a lattice in R^n restricted to a specific box and with
    coordinates in another box.

    INPUT:
    - ``lattice_basis`` -- a list of lists of a basis of the embedded lattice
    - ``lattice_bounds`` -- the bounds for the embeddings
    - ``coordinate_bounds`` -- the bounds for the coordinates
    - ``norm_bound`` -- bounds for the norm (default None - meaning that no bounds are applied)
    
    EXAMPLES::
    
        sage: from hilbert_modgroup.pullback_cython import lattice_elements_in_box
        sage: lattice_elements_in_box([[0,1],[1,0]],[(-1,1),(-1,1)],[(-1,1),(-1,1)])
        [(-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 0), (0, 1), (1, -1), (1, 0), (1, 1)]
        sage: lattice_elements_in_box([[2,0],[0,2]],[(-2,2),(-1,1)],[(-1,1),(-1,1)])
        [(-1, 0), (0, 0), (1, 0)]
        sage: from hilbert_modgroup.all import *
        sage: P1 = HilbertPullback(HilbertModularGroup(5))
        sage: lb, ib = P1._get_lattice_and_ideal_basis()
        sage: lattice_elements_in_box(lb,[(-2,2),(-2,2)],[(-1,1),(-1,1)])
        [(-1, -1), (-1, 0), (0, -1), (0, 0), (0, 1), (1, 0), (1, 1)]
        sage: lattice_elements_in_box(lb,[(-2,2),(-2,2)],[(-1,1),(0,1)])
        [(-1, 0), (0, 0), (0, 1), (1, 0), (1, 1)]
        
    """
    coordinates = integral_coordinates_in_box(coordinate_bounds)
    result = []
    n = len(lattice_basis[0])
    for coordinate_vector in coordinates:
        is_within_bounds = True
        if norm_bound:
            norm = 1
        for i in range(n):
            alpha_i = 0.0
            for j in range(n):
                alpha_i = alpha_i + lattice_basis[i][j] * coordinate_vector[j]
            if alpha_i < lattice_bounds[i][0] or alpha_i > lattice_bounds[i][1]:
                # We need to discard this
                is_within_bounds = False
                break
            if norm_bound:
                norm = norm*alpha_i
        if norm_bound and abs(norm) > norm_bound:
            is_within_bounds = False
        # If we are within the bounds we add the number field element.
        if is_within_bounds:
            result.append(coordinate_vector)
    return result

cpdef coordinates_to_ideal_elements(coordinates,ideal_basis):
    r"""
    Return elements of an ideal given by coordinates and a basis.
    
    INPUT:
    - `coordinates` -- a list of coordinates (as tuples or lists) of the same length as the basis.
    - `ideal_basis` -- the basis of an ideal.
    
    EXAMPLES::
    
        sage: from hilbert_modgroup.pullback_cython import coordinates_to_ideal_elements
        sage: from hilbert_modgroup.all import *
        sage: P1 = HilbertPullback(HilbertModularGroup(5))
        sage: lb, ib = P1._get_lattice_and_ideal_basis()
        sage: coordinates_to_ideal_elements([[1,1],(2,3)],ib)
        [1/2*a + 1/2, 3/2*a + 1/2]
        
    TESTS::
        
        sage: coordinates_to_ideal_elements([[1,1],(2,2,1)],ib)
        Traceback (most recent call last):
        ...
        ValueError: Coordinate need to have same length as basis!
                        
    """
    result = []
    n = len(ideal_basis)
    for coordinate_vector in coordinates:
        if len(coordinate_vector) != n:
            raise ValueError("Coordinate need to have same length as basis!")
        element = 0
        for i,b in enumerate(ideal_basis):
            element += b * coordinate_vector[i]
        result.append(element)
    return result


cpdef find_candidate_cusps(p, z, use_lll=True, use_norm_bound=True, return_sigma_candidates=False,
                           initial_bd_d=None,
                           use_initial_bd_d=True):
    r"""
    Return candidates for closest cusp to the point ``z`` in the upper half-plane. 
    
    INPUT:
    
    - ``p`` -- object of type HilbertPullback
    - ``z`` -- point in the upper half-plane
    - ``use_lll`` -- boolean (default: `True`)  Use the LLL method to find a preliminary bounds
    - ``use_norm_bound`` -- boolean (default: `True`) Use the norm bound together ith the embedding bounds
    - ``return_sigma_candidates`` -- boolean (default: `False`) Return a list of sigma candidates only
    - ``initial_bd_d`` -- positive number (default: `None`) - an initial bound for the distance to nearest cusp.
    - ``use_initial_bd_d`` -- boolean (default: `False`) Use the initial bound
    
    OUTPUT:
    - list of tuples (sigma,rho) representing cusps 
    
    EXAMPLES::
    
        sage: from hilbert_modgroup.all import *
        sage: from hilbert_modgroup.pullback_cython import find_candidate_cusps
        sage: H1 = HilbertModularGroup(5)
        sage: P1 = HilbertPullback(H1)
        sage: z=UpperHalfPlaneProductElement([CC(0,1),CC(0,1)])
        sage: find_candidate_cusps(P1,z)
        [(1, 0), (-1, -1), (0, 1), (1, -1)]
        sage: z=UpperHalfPlaneProductElement([CC(0,0.5),CC(0,1)])
        sage: find_candidate_cusps(P1,z)
        [(0, 1)]        
        sage: H2 = HilbertModularGroup(10)
        sage: P2 = HilbertPullback(H2)
        sage: z=UpperHalfPlaneProductElement([CC(0,1),CC(0,1)])
        sage: find_candidate_cusps(P2,z)
         [(1, 0),
         ...
         sage: len(_)
         10
        sage: z=UpperHalfPlaneProductElement([CC(2.58,0.5),CC(0.5,0.5)]) # long time (1 second)
        sage: len(find_candidate_cusps(P2,z)) # long time (1 second)
        271
        
    """
    ideal_basis = p._construct_ideal(1).integral_basis()
    lattice_basis = p.basis_matrix_ideal()
    n = len(lattice_basis[0])
    # Make lattice basis to a nested list to avoid creation of FreeModule elements
    lattice_basis = [[lattice_basis[i][j] for j in range(n)] for i in range(n)]
    ## Initial bound:
    if use_lll:
        candidate_cusp = p.get_heuristic_closest_cusp(z)
    else:
        candidate_cusp = None
    if candidate_cusp and use_initial_bd_d:
        dist = distance_to_cusp(p, candidate_cusp[0], candidate_cusp[1], z)
    else:
        dist = None
    if use_norm_bound:
        norm_bound_sigma = p._bound_for_sigma_norm(z, dist)
        norm_bound_sigma = norm_bound_sigma*(1+2**(-z[0].prec()/2)) # correct for numerical errors
    else:
        norm_bound_sigma = None
    if use_initial_bd_d and initial_bd_d and (not dist or dist > initial_bd_d):
        dist = initial_bd_d
        # norm_bound_rho = p._bound_for_rho_norm(z, dist)
    one = p.number_field()(1)
    coordinate_bounds = p._bound_for_sigma_coordinates(z, initial_bd_d=dist,use_initial_bd_d=use_initial_bd_d)
    embedding_bounds = p._bound_for_sigma_embeddings(z, initial_bd_d=dist,use_initial_bd_d=use_initial_bd_d)
    coordinate_bounds = [(-b,b) for b in coordinate_bounds]
    embedding_bounds = [(-b,b) for b in embedding_bounds]
    sigma_candidates_coordinates = lattice_elements_in_box(lattice_basis,
                                               embedding_bounds,
                                               coordinate_bounds, norm_bound=norm_bound_sigma)
    sigma_candidates = coordinates_to_ideal_elements(sigma_candidates_coordinates,
                                                             ideal_basis)
    if return_sigma_candidates:
        return sigma_candidates
    # To remove duplicates we keep track of the quotients rho/sigma (for sigma !=0)
    quotients = {}
    if not candidate_cusp or candidate_cusp[1] == 0:
        result = [(p.number_field()(1), p.number_field()(0))]
    elif candidate_cusp[0] == 0:
        result = [(p.number_field()(0), p.number_field()(1))]
    else:
        # We always have infinity since sigma=0 is always within the bounds.
        result = [candidate_cusp,(p.number_field()(1), p.number_field()(0))]
        q = candidate_cusp[0]/candidate_cusp[1]
        ngcd = gcd(candidate_cusp[0].norm(),candidate_cusp[1].norm())
        quotients = {q: (ngcd,(candidate_cusp[0],candidate_cusp[1]))}
    for s in sigma_candidates:
        if s == 0:
            continue
        rho_coordinate_bounds = p._bound_for_rho_coordinates(z, s, initial_bd_d=dist, use_initial_bd_d=use_initial_bd_d)
        rho_coordinate_bounds = [(-b,b) for b in rho_coordinate_bounds]
        rho_embedding_bounds = p._bound_for_rho_embeddings(z, s, initial_bd_d=dist, use_initial_bd_d=use_initial_bd_d)

        rho_candidates_coordinates = lattice_elements_in_box(lattice_basis,
                                                 rho_embedding_bounds,
                                                 rho_coordinate_bounds)
        rho_candidates = coordinates_to_ideal_elements(rho_candidates_coordinates,
                                                         ideal_basis)

        for r in rho_candidates:
            if r == 0 and (r,one) not in result:
                result.append((r,one))
            if r == 0:
                continue
            quo = r/s
            ngcd = gcd(r.norm(), s.norm())
            if quo in quotients:
                # We try to pick the representative cusp that has the smallest norm gcd
                if ngcd < quotients[quo][0]:
                    result.remove(quotients[quo][1])
                else:
                    continue
            result.append((r,s))
            quotients[quo] = (ngcd,(r,s))
    return result

cpdef find_closest_cusp(p, z, return_multiple=False, use_lll=True, use_norm_bound=True):
    r"""
    Return the closest cusp(s) to z
    
    INPUT:
    - ``p`` -- A HilbertPullback object
    - ``z`` -- point in the upper half-plane
    - ``return_multiple`` -- boolean (default False), set to True to return multiple closest cusp if it is no unique. 
    - ``use_lll`` -- boolean (default True) if True uses the LLL method to obtain an initial bound
    - ``use_norm_bound`` -- boolean (default True) if True uses the norm bound as well as the embedding bounds

    OUTPUT:
    - tuple representing a cusp (if return_multiple == False)
    - list of tuples representing cusps (if return_multiple == True) 
    
    EXAMPLES::

        sage: from hilbert_modgroup.all import *
        sage: from hilbert_modgroup.pullback_cython import find_closest_cusp
        sage: H1 = HilbertModularGroup(5)
        sage: P1 = HilbertPullback(H1)
        sage: z=UpperHalfPlaneProductElement([CC(0,1),CC(0,1)])
        sage: find_closest_cusp(P1,z)
        (1, 0) 
        sage: find_closest_cusp(P1,z,return_multiple=True)
        [(1, 0), (0, 1)]
        sage: z=UpperHalfPlaneProductElement([CC(0,0.5),CC(0,1)])
        sage: find_closest_cusp(P1,z)
        (0, 1)
        sage: find_closest_cusp(P1,z,return_multiple=True)
        [(0, 1)]
        
        sage: H2 = HilbertModularGroup(10)
        sage: P2 = HilbertPullback(H2)
        sage: z=UpperHalfPlaneProductElement([CC(0,1),CC(0,1)])
        sage: find_closest_cusp(P2,z)
        (1, 0)
        sage: find_closest_cusp(P2,z,return_multiple=True)
        [(1, 0), (0, 1)]
        sage: z=UpperHalfPlaneProductElement([CC(-2.58,0.5),CC(0.5,0.5)])
        sage: find_closest_cusp(P2,z,return_multiple=True)
        [(-a + 2, -2)]
        
    """
    cusp_candidates = find_candidate_cusps(p, z, use_lll=use_lll, use_norm_bound=use_norm_bound)
    min_cusp = cusp_candidates[0]
    min_d = distance_to_cusp(p,min_cusp[0], min_cusp[1], z)
    if return_multiple:
        min_cusp = [min_cusp]
    min_cusp_bound = p._bound_for_closest_cusp()
    eps = 2.0**(4-53) # Numerical precision for when we consider cusps to be equally close.
    # There may be many identical cusps in this list.
    # In particular there might be many
    for c in cusp_candidates[1:]:
        d = distance_to_cusp(p, c[0],c[1], z)
        if abs(d-min_d)<eps and return_multiple:
            min_cusp.append(c)
        if d < min_d - 2*eps:
            if return_multiple:
                min_cusp = [c]
            else:
                min_cusp = c
            min_d = d
        if d < min_cusp_bound:
            break
    # We have already filtered away identical cusps but as a final stage
    # we also check if the minimal cusps are equal to any of the fixed representatives.
    # other than infinity which is already taken care of
    if p.group().ncusps() == 1:
        return min_cusp
    result = []
    for cusp in p.group().cusps()[1:]:
        c, d = cusp.numerator(),cusp.denominator()
        quo = c/d
        if return_multiple:
            for r,s in min_cusp:
                if s == 0:
                    result.append((r, s))
                elif r == 0 and (0,1) not in result:
                    result.append((0,1))
                elif r == 0:
                    continue
                elif r/s == quo and (c,d) not in result:
                    result.append((c,d))
                elif r/s != quo:
                    result.append((r,s))
        elif min_cusp[1] != 0 and min_cusp[0]/min_cusp[1] == quo:
            result = (c,d)
        else:
            result = min_cusp
    return result

cpdef distance_to_cusp(SageObject p, NumberFieldElement r, NumberFieldElement s,
                        UpperHalfPlaneProductElement__class z):
    r"""
    Return the distance from an element in the upper half-plane to a cusp. 
    
    INPUT:
    - ``p`` - Should be an object of type HilbertPullback (but this is not recognized by Cython so I use SageObject instead)
    - ``r`` - Number field element, numerator of cusp
    - ``s`` - Number field element, denomnator of cusp
    - ``z`` - Point in the upper half-plane
    
    EXAMPLES::
    
        sage: from hilbert_modgroup.all import *
        sage: from hilbert_modgroup.pullback_cython import distance_to_cusp
        sage: H1 = HilbertModularGroup(5)
        sage: P1 = HilbertPullback(H1)
        sage: K = P1.number_field()
        sage: z=UpperHalfPlaneProductElement([CC(0,1),CC(0,1)])
        sage: distance_to_cusp(P1,K(0),K(1),z) # abs tol 1e-10
        1.0
        sage: distance_to_cusp(P1,K(1),K(1),z) # abs tol 1e-10
        2.0
        sage: z=UpperHalfPlaneProductElement([CC(0,0.5),CC(0,1)])
        sage: distance_to_cusp(P1,K(0),K(1),z) # abs tol 1e-10
        0.707106781186548        
        sage: H2 = HilbertModularGroup(10)
        sage: P2 = HilbertPullback(H2)
        sage: K = P2.number_field()
        sage: z=UpperHalfPlaneProductElement([CC(0,1),CC(0,1)])
        sage: distance_to_cusp(P2,K(0),K(1),z) # abs tol 1e-10
        1.0
        sage: z=UpperHalfPlaneProductElement([CC(-2.58,0.5),CC(0.5,0.5)])
        sage: a = K.gen()
        sage: distance_to_cusp(P2,a-5,-a,z) # abs tol 1e-10
        1.01308408502928
 
    """
    cdef list rlist,slist
    ideal_rs = r.parent().fractional_ideal(r,s)
    rlist = r.complex_embeddings()
    slist = s.complex_embeddings()
    n = len(slist)
    d = 1
    for i in range(n):
        d = d* ((-z._x[i] * slist[i] + rlist[i]) ** 2 * z._y[i] ** -1 + slist[i] ** 2 * z._y[i])
    d = ideal_rs.norm()**-1*d.sqrt()
    return d


"""
General utility functions

"""
from sage.rings.real_mpfi import RealIntervalField
from sage.structure.element import Vector


def upper(x, prec):
    r"""
    Return an upper bound to x given by precision prec.

    EXAMPLES::

        sage: from hilbert_modgroup.utils import upper
        sage: upper(0.1,8)
        0.11
        sage: upper(0.1,53)
        0.100000000000001
        sage: upper(RR.pi(),16)
        3.142
    """
    return RealIntervalField(prec)(x).upper()


def lower(x, prec):
    r"""
    Return a lower bound to x given by precision prec.

    EXAMPLES::

        sage: from hilbert_modgroup.utils import lower
        sage: lower(0.1,8)
        0.099
        sage: lower(0.1,30)
        0.099999999
        sage: lower(RR.pi(),16)
        3.141
    """
    return RealIntervalField(prec)(x).lower()


def get_prec(x, **kwds):
    r"""
    Return the precision of x.

    EXAMPLES::

        sage: from hilbert_modgroup.utils import get_prec
        sage: get_prec(0.1)
        53
        sage: get_prec(RR.pi())
        53
        sage: get_prec([RealField(103).pi(), RealField(103)(1)])
        103
    """
    if 'prec' in kwds:
        return kwds['prec']
    try:
        if isinstance(x, (list, tuple, Vector)):
            prec = x[0].prec()
        else:
            prec = x.prec()
    except AttributeError:
        prec = 53
    return prec



import os
import shutil
import subprocess
import setuptools
import sysconfig
from setuptools.extension import Extension
from Cython.Build import cythonize
# Find correct value for SAGE_LIB which is needed to compile the Cython extensions.
try:
    from sage.env import SAGE_LIB
except ModuleNotFoundError:
    SAGE_LIB = os.getenv('SAGE_LIB') or sysconfig.get_path('purelib')
# Check if sage is installed in SAGE_LIB
if not os.path.isdir(SAGE_LIB + '/sage'):
    raise ModuleNotFoundError("Sagemath or passagemath needs to be installed.")

INCLUDE_DIRS = []
LIBRARY_DIRS = []
if shutil.which('brew') is not None:
    proc = subprocess.Popen("brew --prefix", shell=True,
                            stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                            stderr=subprocess.STDOUT, close_fds=True)
    HOMEBREW_PREFIX = proc.stdout.readline().decode('utf-8').strip()
    HOMEBREW_LIB = HOMEBREW_PREFIX + '/lib'
    LIBRARY_DIRS.append(HOMEBREW_LIB)
    HOMEBREW_INC = HOMEBREW_PREFIX + '/include'
    INCLUDE_DIRS.append(HOMEBREW_INC)

# Extension modules using Cython
extra_compile_args = ['-Wno-unused-function',
                      '-Wno-implicit-function-declaration',
                      '-Wno-unused-variable',
                      '-Wno-deprecated-declarations',
                      '-Wno-deprecated-register']
ext_modules = [
    Extension(
        'hilbert_modgroup.hilbert_modular_group_element',
        sources=[os.path.join('src/hilbert_modgroup/hilbert_modular_group_element.pyx')],
        extra_compile_args=extra_compile_args,
        include_dirs=INCLUDE_DIRS, library_dirs=LIBRARY_DIRS
    ),
    Extension(
        'hilbert_modgroup.upper_half_plane',
        sources=[os.path.join('src/hilbert_modgroup/upper_half_plane.pyx')],
        extra_compile_args=extra_compile_args,
        include_dirs=INCLUDE_DIRS, library_dirs=LIBRARY_DIRS
    ),
    Extension(
        'hilbert_modgroup.pullback_cython',
        sources=[os.path.join('src/hilbert_modgroup/pullback_cython.pyx')],
        language='c++',
        extra_compile_args=extra_compile_args+['-std=c++11'],
        include_dirs=INCLUDE_DIRS, library_dirs=LIBRARY_DIRS
    )
]

from contextlib import contextmanager

try:
    from sage.misc.package_dir import cython_namespace_package_support
except ImportError:
    @contextmanager
    def cython_namespace_package_support():
        yield

with cython_namespace_package_support():
    setuptools.setup(
        ext_modules=cythonize(
            ext_modules,
            include_path=['src', SAGE_LIB],
            compiler_directives={
                'embedsignature': True,
                'language_level': '3',
            },
        ),
    )
